#include <Arduino.h>

#include "ntp.h"
#include "config.h"
#include "my_wifi.h"
#include "ota.h"
#include "my_fastled.h"
#include "my_mqtt.h"
#include "functions.h"
#include "effects.h"
#include "http_server.h"
#include "recorder.h"

uint8_t starting_up = OTA_STARTUP_DELAY;
int loop_timeouts = 0;
long loop_started_at = 0;
uint8_t baseHue = 0; // defined as extern in prototypes.h
char hostname[30]; // defined as extern in prototypes.h
#ifdef RECORDER_ENABLE
Recorder* recorder;
#endif
#ifdef MQTT_REPORT_METRICS
uint16_t metrics_frame_count = 0;
unsigned long metrics_frame_time = 0;
#endif

void setup() {
    Serial.begin(74880);
    LOGln("Core * Starting");

	int chipid;
	#if defined( ESP8266 )
	chipid = ESP.getChipId();
	#elif defined( ESP32 )
	chipid = ESP.getEfuseMac() & 0xFFFFFF;
	#endif
	snprintf(hostname, 30, HOSTNAME, chipid);

    setup_effects();
    wifi_setup();
    ntp_setup();
    ota_setup();
    fastled_setup();
    ntpClient.begin();
    #ifdef HTTP_SERVER_ENABLE
    http_server_setup();
    #endif
    #ifdef MQTT_ENABLE
    mqtt_setup();
    #endif
    #ifdef RECORDER_ENABLE
	recorder = new Recorder();
	#endif
    SPIFFS.begin();
    LOGln("Core * Setup complete");
}

void loop() {
    loop_started_at = millis();
    ota_loop();

    if (starting_up > 0) {
        EVERY_N_SECONDS(1) {
            Serial.print("Core * Waiting for OTA... "); Serial.println(starting_up);
            starting_up--;
            Window* w = Window::getFullWindow();
            CRGB color(0xFF0000);
            w->clear();
            for (int i=0; i<starting_up; i++) {
                w->setPixel(i, 0, &color);
            }
            FastLED.show();
        }
        return;
    }

    ntpClient.update();
    #ifdef MQTT_ENABLE
    mqtt_loop();
    #endif
    #ifdef HTTP_SERVER_ENABLE
    http_server_loop();
    #endif

    EVERY_N_MILLISECONDS(100) {
        baseHue++;
    }

    EVERY_N_MILLISECONDS(1000 / FPS) {
        //LOGln("Core * loop running");
        current_effect->loop();
        //LOGln("Core * loop ran");

        if (current_effect->can_be_shown_with_clock()) {
            effect_clock.loop(current_effect->clock_as_mask(), CRGB(0xFFFFFF), CRGB(0x000000));
        }
        FastLED.show();
        #ifdef MQTT_REPORT_METRICS
        metrics_frame_count++;
        metrics_frame_time += (millis() - loop_started_at);
        #endif
        
        #ifdef RECORDER_ENABLE
        recorder->loop();
        #endif
    }

    #if defined(MQTT_ENABLE) && defined(MQTT_REPORT_METRICS)
    EVERY_N_SECONDS(15) {
        char json[120];
        snprintf(json, 120, "{\"free_heap\":%u, \"uptime\":%lu, \"fps\":%d, \"time_per_frame\":%lu}", ESP.getFreeHeap(), millis()/1000, FastLED.getFPS(), metrics_frame_time / metrics_frame_count);
        mqtt_publish("metrics", json);
        metrics_frame_count = 0;
        metrics_frame_time = 0;
    }
    #endif // MQTT_REPORT_METRICS


    if (MONITOR_LOOP_TIMES && millis()-loop_started_at>=MONITOR_LOOP_TIME_THRESHOLD) {
        //LOG("Core * Loop took "); LOG(millis()-loop_started_at); LOGln("ms.");
        loop_timeouts++;
        //LOG("Core * Timeout counter is now "); LOGln(loop_timeouts);
        if (loop_timeouts >= MONITOR_LOOP_TIME_COUNT_MAX) {
            ESP.restart();
        }
    } else if (loop_timeouts > 0) {
        loop_timeouts--;
    }
}
